/*
 *  Annealing.cpp
 *  alife-xcode30
 *
 *  Created by Eduardo Poyart on 2/22/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#include "alife_Prefix.pch"
#include "common.h"

#define CHANGE_DISTANCE_SNOWBOARDER 0.5
#define CHANGE_DISTANCE_SPIDER 0.05
#define TIMESTEP (1.0 / 60.0)
#define FINAL_TEMP 0.001
#define TEMP_CHANGE 0.9998
#define NUM_RANDOM_CHANGES 2

void Annealing::init()
{
	m_ofs.open("annealing.out", ios::out);
	if (!m_ofs.good())
	{
		cout << "Error opening annealing.out" << endl;
		exit(1);
	}
}

void Annealing::random_change()
{
	if (g_params.model == MODEL_SNOWBOARDER)
	{
		real rfactor = 2.0 * CHANGE_DISTANCE_SNOWBOARDER * (1.0 * (m_t + 1.0));
		
		for (int i = 0; i < NUM_RANDOM_CHANGES; i++)
		{
			real r = ((real)rand() / (real)RAND_MAX) - 0.5;
			r *= rfactor;
			int index = rand() % m_nn->parameter_space();
			real* w = m_nn->parameter(index);		
			*w += r;
		}
	}
	else if (g_params.model == MODEL_SPIDER)
	{
		real rfactor = 2.0 * CHANGE_DISTANCE_SPIDER * (1.0 * (m_t + 1.0));
		
		for (int i = 0; i < NUM_RANDOM_CHANGES; i++)
		{
			real r = ((real)rand() / (real)RAND_MAX) - 0.5;
			r *= rfactor;
			int index = rand() % m_fj->parameter_space();
			real* w = m_fj->parameter(index);		
			*w += r;
		}
	}
	else if (g_params.model == MODEL_CYCLIST)
	{
		real rfactor = 2.0 * CHANGE_DISTANCE_SNOWBOARDER * (1.0 * (m_t + 1.0));
		
		for (int i = 0; i < NUM_RANDOM_CHANGES; i++)
		{
			real r = ((real)rand() / (real)RAND_MAX) - 0.5;
			r *= rfactor;
			int index = rand() % m_nn->parameter_space();
			real* w = m_nn->parameter(index);		
			*w += r;
		}
	}
	else
		assert(0);
}

real Annealing::fitness_snowboarder()
{
	Vector3 pos;
	m_model->get_head_pos(pos.get_v());
	return pos[0]; // * 20 + m_step_counter;
}

real Annealing::fitness_spider()
{
	Vector3 pos;
	m_model->get_head_pos(pos.get_v());
	return pos[1];
}

real Annealing::fitness_cyclist()
{
	return fitness_snowboarder();
}

void Annealing::save_and_exit()
{
	if (g_params.model == MODEL_SNOWBOARDER || g_params.model == MODEL_CYCLIST)
		m_nn->save("nn.txt");
	else if (g_params.model == MODEL_SPIDER)
		m_fj->save("fj.txt");
	else
		assert(0);
	cout << "Done." << endl;
	m_ofs.close();
	exit(0);	
}

bool Annealing::force_stop()
{
	bool stop;
	ifstream ifs;
	ifs.open("stop.txt", ios::in);
	stop = (ifs.good());
	ifs.close();
	return stop;
}

void Annealing::step()
{
	if (m_stop)
		save_and_exit();
	
	bool end_condition;
	end_condition = (m_time > m_total_time) || m_abort;

	real fitness;
	if (g_params.model == MODEL_SNOWBOARDER)
		fitness = fitness_snowboarder();
	else if (g_params.model == MODEL_SPIDER)
		fitness = fitness_spider();
	else if (g_params.model == MODEL_CYCLIST)
		fitness = fitness_cyclist();
	
	if (end_condition)
	{
		m_abort = false;
		bool rejected = false;
		m_step_counter = 0;
		m_time = 0.0;
		m_t *= TEMP_CHANGE;
		m_cycle_count++;
		cout << "Cycle " << m_cycle_count << " t=" << m_t << " bf=" << m_best_fitness << " f=" << fitness;
		m_ofs << m_cycle_count << ", " << m_t << ", " << m_best_fitness << ", " << fitness << endl;
		if (fitness > m_best_fitness)
		{
			m_best_fitness = fitness;
			cout << " improved" << endl;
		}
		else
		{
			if (!should_change_state())
			{
				if (g_params.model == MODEL_SNOWBOARDER || g_params.model == MODEL_CYCLIST)
					*m_nn = m_prev_nn;
				else if (g_params.model == MODEL_SPIDER)
					*m_fj = m_prev_fj;
				rejected = true;
				cout << " rejected" << endl;
			}
			else
			{
				m_best_fitness = fitness;
				cout << " reduced" << endl;
			}
		}
		if (m_t < FINAL_TEMP || force_stop() )
			save_and_exit();
		
		m_model->reset();
		if (!rejected)
			if (g_params.model == MODEL_SNOWBOARDER || g_params.model == MODEL_CYCLIST)
				m_prev_nn = *m_nn;
			else if (g_params.model == MODEL_SPIDER)
				m_prev_fj = *m_fj;
		random_change();
	}
	m_sim->step(TIMESTEP);
	m_time += TIMESTEP;
	m_step_counter++;
}
