/*
 *  Timer.h
 *  alife-xcode30
 *
 *  Created by Eduardo Poyart on 2/22/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#if RENDER

#ifdef LINUX

#include <sys/time.h>

class Timer
{
protected:
	timespec ts;

public:
	Timer() 
	{
		clock_gettime(CLOCK_REALTIME, &ts);
	}
	virtual ~Timer() {}
	real elapsed()
	{
		timespec cts;
		clock_gettime(CLOCK_REALTIME, &cts);
		time_t sec_diff = cts.tv_sec - ts.tv_sec;
		real result = (real)sec_diff + ((real)cts.tv_nsec - (real)ts.tv_nsec) * 1.0e-9;
		ts.tv_sec = cts.tv_sec;
		ts.tv_nsec = cts.tv_nsec;
		return result;
	}
};

#else
class Timer
{
protected:
	uint64_t m_time;
	real m_timebase_info;

public:
	Timer(): m_time(mach_absolute_time()) 
	{
		mach_timebase_info_data_t tbinfo;	
		mach_timebase_info(&tbinfo);
		m_timebase_info = (real)(1.0e-9 * (double)tbinfo.numer / (double)tbinfo.denom);
	}
	virtual ~Timer() {}
	real elapsed()
	{
		uint64_t current_time = mach_absolute_time();
		uint64_t delta = current_time - m_time;
		m_time = current_time;
		return (real)delta * m_timebase_info;
	}
};

#endif

#else
class Timer
{
public:
	Timer() {}
	virtual ~Timer() {}
	real elapsed() { return 0.0; }
};	
#endif

