/*
 *  Model.h
 *  alife
 *
 *  Created by Eduardo Poyart on 2/17/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

enum SolidType { SOLID_BOX, SOLID_CYLINDER, SOLID_SPHERE };

struct Solid
{
	string name;
	SolidType type;
	Vector3 translation;
	Vector3 scale;
	Vector4 rotation;
	dBodyID bodyid;
	dGeomID geomid;
};

enum JointType { JOINT_UNIVERSAL, JOINT_HINGE, JOINT_BALL };

struct Joint
{
	string name;
	JointType type;
	Vector3 translation;
	Vector4 rotation;
	string bodyname[2];
	dBodyID bodyid[2];
	bool has_limit[2];
	Vector2 limit[2];
	dJointID jointid;
	Joint() { has_limit[0] = has_limit[1] = false; }
};

class Model
{
public:
	Model(): m_nn(NULL), m_fj(NULL) {}
	virtual ~Model() {}
	
	void load(const string& filename, dWorldID world, dSpaceID space);
	void step(real t);
	void draw() const;
	void reset();
	void get_head_pos(real* pos) const;
	//void set_pos(const real* pos);
	int get_joint_by_name(const string& name);
	dBodyID get_bodyid_by_name(const string& name);
	int get_solid_index_by_name(const string& name);
	dGeomID get_geomid_by_name(const string& name);
	void replace_geomid_by_name(const string& name, dGeomID geom_id);
	dBodyID get_head_bodyid() { return m_head_bodyid; }
	dBodyID get_board_bodyid() { return m_board_bodyid; }
	dBodyID get_bicycle_bodyid() { return m_bicycle_bodyid; }
	dBodyID get_bicycle_wheel1id() { return m_bicycle_wheel1id; }
	dBodyID get_bicycle_wheel2id() { return m_bicycle_wheel2id; }
	NeuralNet* get_nn() { return m_nn; }
	FourierJoint* get_fj() { return m_fj; }
	vector<Solid>& get_solids() { return m_solids; }
	vector<Solid>& get_geoms() { return m_geoms; }
	void add_geom(Solid& s, dSpaceID space);
	
protected:
	vector<Solid> m_solids;
	vector<Joint> m_joints;
	vector<Solid> m_geoms;
	vector<dBodyID> m_input_bodies;
	vector<int> m_input_joints;
	vector<int> m_control_joints;
	vector<int> m_control_feet_joints;
	vector<int> m_spring_joints;
	dBodyID m_head_bodyid;
	dBodyID m_board_bodyid;
	dBodyID m_bicycle_bodyid;
	dBodyID m_bicycle_wheel1id;
	dBodyID m_bicycle_wheel2id;
	real m_total_time;
	
	NeuralNet* m_nn;
	vector<real> m_nn_in;
	vector<real> m_nn_out;
	
	FourierJoint* m_fj;
#if RENDER
	GLUquadricObj* m_sphereobj;
#endif
	
	void load_solid(Parser& p);
	void load_joint(Parser& p, Joint& j);
	void generate_ground();
	void step_snowboarder(real t);
	void step_spider(real t);
	void step_cyclist(real t);
	void set_pd_universal(const Joint& jnt, real val1, real val2, real ks, real kd);
	void set_pd_hinge(const Joint& jnt, real val, real ks, real kd);
	void set_spring_universal(const Joint& jnt, real val1, real val2, real ks, real kd);
	void set_spring_hinge(const Joint& jnt, real val, real ks, real kd);
	void draw_solid(const Solid& s) const;
	void draw_box(const real sides[3]) const;
	void draw_cylinder(float l, float r, float zoffset) const;
	void draw_sphere(const real scale[3]) const;
	void create_geom(Solid& s, dSpaceID space);
};
