/*
 *  NeuralNet.cpp
 *  alife-xcode30
 *
 *  Created by Eduardo Poyart on 2/22/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#include "alife_Prefix.pch"
#include "common.h"

#define DEBUG_IN_OUT 0

void NeuralNet::compute(const vector<real>& input, vector<real>& output)
{
	assert(input.size() == m_n_in);
	assert(output.size() == m_n_out);
	
	memset(m_vals.a, 0, m_vals.m * m_vals.n * sizeof(real));
	
#if DEBUG_IN_OUT
	// Debug
	for (int i = 0; i < m_n_in; i++)
		cout << input[i] << " ";
	cout << endl;
#endif
	
	// Init
	for (int i = 0; i < m_n_in; i++)
		m_vals(0, i) = input[i];
	
	// Zero unused values
	for (int i = m_n_in; i < m_n_hnodes; i++)
		for (int j = 0; j < m_n_hnodes; j++)
			m_weights(0, i, j) = 0.0;
		
	// Compute
	register real v;
	for (int i = 0; i < m_n_hlayers + 1; i++)
	{
		for (int j = 0; j < m_n_hnodes; j++)
		{
			v = m_vals(i+1, j);
			for (int k = 0; k < m_n_hnodes; k++)
				v += m_vals(i, k) * m_weights(i, k, j);
			m_vals(i+1, j) = sigmoid(v + m_offsets(i, j));
			//cout << m_vals(i+1, j) << " ";
		}
		//cout << endl;
	}
	
	// Output
	for (int i = 0; i < m_n_out; i++)
		output[i] = m_vals(m_n_hlayers+1, i);

#if DEBUG_IN_OUT
	// Debug
	for (int i = 0; i < m_n_out; i++)
		cout << output[i] << " ";
	cout << endl;
#endif
}

void NeuralNet::save(const char* filename)
{
	ofstream ofs;
	ofs.open(filename, ios::out);
	if (!ofs.good())
	{
		cout << "Error opening output file." << endl;
		exit(1);
	}
	ofs << m_weights.m << " " << m_weights.n << " " << m_weights.p << endl;
	for (int i = 0; i < m_weights.m; i++)
		for (int j = 0; j < m_weights.n; j++)
			for (int k = 0; k < m_weights.p; k++)
				ofs << m_weights(i, j, k) << endl;
	ofs << m_offsets.m << " " << m_offsets.n << endl;
	for (int i = 0; i < m_offsets.m; i++)
		for (int j = 0; j < m_offsets.n; j++)
			ofs << m_offsets(i, j) << endl;
	ofs.close();
}

void NeuralNet::load(const char* filename)
{
	ifstream ifs;
	ifs.open(filename, ios::in);
	if (!ifs.good())
	{
		cout << "Error opening input file." << endl;
		exit(1);
	}
	int m, n, p;
	ifs >> m >> n >> p;
	m_weights.resize(m, n, p);
	for (int i = 0; i < m_weights.m; i++)
		for (int j = 0; j < m_weights.n; j++)
			for (int k = 0; k < m_weights.p; k++)
				ifs >> m_weights(i, j, k);
	ifs >> m >> n;
	m_offsets.resize(m, n);
	for (int i = 0; i < m_offsets.m; i++)
		for (int j = 0; j < m_offsets.n; j++)
			ifs >> m_offsets(i, j);
	ifs.close();
}
