/*
 *  NeuralNet.h
 *  alife-xcode30
 *
 *  Created by Eduardo Poyart on 2/22/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#define RAND1 (((real)rand() / RAND_MAX) * 2.0 - 1.0)

template<class T>
struct Array2d
{
	Array2d(int in_m, int in_n): m(in_m), n(in_n) { a = new T[in_m * in_n]; }
	Array2d(const Array2d& ar): m(ar.m), n(ar.n) { a = new T[ar.m * ar.n]; memcpy(a, ar.a, sizeof(T) * ar.m * ar.n); }
	virtual ~Array2d() { delete[] a; }
	const Array2d& operator=(const Array2d& ar) { assert(this != &ar); delete[] a; m = ar.m; n = ar.n;
		a = new T[m * n]; memcpy(a, ar.a, sizeof(T) * m * n); return *this; }
	T& operator()(int x, int y) { assert(x >= 0 && x < m && y >= 0 && y < n); return a[x*n + y]; }
	void resize(int in_m, int in_n) { m=in_m; n=in_n; delete[] a; a = new T[in_m * in_n]; }
	int m, n;
	T* a;
};

template<class T>
struct Array3d
{
	Array3d(int in_m, int in_n, int in_p): m(in_m), n(in_n), p(in_p) { a = new T[in_m * in_n * in_p]; }
	Array3d(const Array3d& ar): m(ar.m), n(ar.n), p(ar.p) { a = new T[ar.m * ar.n * ar.p]; memcpy(a, ar.a, sizeof(T) * ar.m * ar.n * ar.p); }
	virtual ~Array3d() { delete[] a; }
	const Array3d& operator=(const Array3d& ar) { assert(this != &ar); delete[] a; m = ar.m; n = ar.n; p = ar.p;
		a = new T[m * n * p]; memcpy(a, ar.a, sizeof(T) * m * n * p); return *this; }
	T& operator()(int x, int y, int z) { assert(x >= 0 && x < m && y >= 0 && y < n && z >= 0 && z < p); return a[(x*n + y)*p + z]; }
	void resize(int in_m, int in_n, int in_p) { m=in_m; n=in_n; p=in_p; delete[] a; a = new T[in_m * in_n * in_p]; }
	int m, n, p;
	T* a;
};

class NeuralNet
{
protected:
	int m_n_in;
	int m_n_hlayers;
	int m_n_hnodes;
	int m_n_out;
	Array3d<real> m_weights;
	Array2d<real> m_vals;
	Array2d<real> m_offsets;

public:
	NeuralNet(int n_inputs, int n_hlayers, int n_hnodes, int n_outputs):
		m_n_in(n_inputs), m_n_hlayers(n_hlayers), m_n_hnodes(n_hnodes), m_n_out(n_outputs),
		m_weights(n_hlayers + 1, n_hnodes, n_hnodes), m_vals(n_hlayers + 2, n_hnodes), m_offsets(n_hlayers + 1, n_hnodes)
	{
		//random_init();
		memset(m_weights.a, 0, (n_hlayers + 1) * n_hnodes * n_hnodes * sizeof(real));
		memset(m_offsets.a, 0, (n_hlayers + 1) * n_hnodes * sizeof(real));
	}
	virtual ~NeuralNet() {}
	NeuralNet(const NeuralNet& nn):
		m_n_in(nn.m_n_in), m_n_hlayers(nn.m_n_hlayers), m_n_hnodes(nn.m_n_hnodes), m_n_out(nn.m_n_out),
		m_weights(nn.m_weights), m_vals(nn.m_vals), m_offsets(nn.m_offsets)
	{ }
	
	const NeuralNet& operator=(const NeuralNet& nn) {
		assert(this != &nn);
		m_n_in = nn.m_n_in; m_n_out = nn.m_n_out; m_n_hlayers = nn.m_n_hlayers; m_n_hnodes = nn.m_n_hnodes;
		m_weights = nn.m_weights; m_vals = nn.m_vals; m_offsets = nn.m_offsets; return *this;
	}
	
	void random_init()
	{
		for (int i = 0; i < m_weights.m; i++)
			for (int j = 0; j < m_weights.n; j++)
				for (int k = 0; k < m_weights.p; k++)
					m_weights(i,j,k) = RAND1;
		for (int i = 0; i < m_offsets.m; i++)
			for (int j = 0; j < m_offsets.n; j++)
				m_offsets(i,j) = RAND1;
	}
	
	int parameter_space()
	{
		return m_weights.m * m_weights.n * m_weights.p + m_offsets.m * m_offsets.n;
	}
	
	real* parameter(int n)
	{
		int num_w = m_weights.m * m_weights.n * m_weights.p;
		int num_o = m_offsets.m * m_offsets.n;
		assert(n >= 0 && n < num_w + num_o);
		if (n >= num_w)
			return &(m_offsets.a[n - num_w]);
		return &(m_weights.a[n]);
	}
	
	void compute(const vector<real>& input, vector<real>& output);
	void save(const char* filename);
	void load(const char* filename);
	
protected:
	inline real sigmoid(real x)
	{
		//real y = x1 / sqrtf(1.0f + x1 * x1);
		return (2.0 / (1.0 + exp(-x))) - 1.0;
	}
};
