/*
 *  Simulator.cpp
 *  alife
 *
 *  Created by Eduardo Poyart on 2/17/10.
 *  Copyright 2010 __MyCompanyName__. All rights reserved.
 *
 */

#include "alife_Prefix.pch"
#include "common.h"

#define COLLIDE_WITH_SELF 0

extern Annealing* the_annealing;

void Simulator::init()
{
	m_world = dWorldCreate();
	m_contactgroup = dJointGroupCreate(0);
	dWorldSetGravity(m_world, 0.0, 0.0, -9.8);

	m_space = dHashSpaceCreate(0);
	if (g_params.model == MODEL_BOARD)
	{
		Vector3 v(0.6, 0.0, 1.0);
		v.normalize();
		m_ground.push_back(dCreatePlane(m_space, v.x, v.y, v.z, 0.0));
	}
	else if (g_params.model == MODEL_SPIDER)
		m_ground.push_back(dCreatePlane(m_space, 0.0, 0.0, 1.0, 0.0));
	dSpaceSetCleanup(m_space, 0);
	//dWorldSetAutoDisableFlag(m_world, 1);
	//dWorldSetERP(m_world, 0.1);
	//dWorldSetCFM(m_world, 0.001);
	
	//m_model.init_test(m_world, m_space);
	if (g_params.model == MODEL_SNOWBOARDER || g_params.model == MODEL_CYCLIST)
	{
		if (g_params.model == MODEL_SNOWBOARDER)
			m_model.load("snowboarder.x3d", m_world, m_space);
		else
			m_model.load("cyclist.x3d", m_world, m_space);
		m_rot_x = 30.0;
		m_rot_z = 0.0;
		m_cam_dist = 15.0;
		for (int i = 0; i < m_model.get_geoms().size(); i++)
			m_ground.push_back(m_model.get_geoms()[i].geomid);
		if (g_params.model == MODEL_SNOWBOARDER)
			generate_ground(Vector3(21.0, 0.0, -9.5), Vector4(-0.15, 0.0, 1.0, 0.0), Vector3(50.0, 25.0, 5.0), 0.2, -7.5);
		else
			generate_ground(Vector3(21.0, 0.0, -7.0), Vector4(-0.1, 0.0, 1.0, 0.0), Vector3(50.0, 25.0, 5.0), 0.2, -7.5);
		//m_mesh.init(m_space);
		//m_ground = m_mesh.get_geom();
		/*
		// Board extra collision
		int boardidx = m_model.get_solid_index_by_name("Board");
		vector<Solid>& solids = m_model.get_solids();
		Solid& s = solids[boardidx];
		dGeomDestroy(s.geomid);
		s.geomid = dCreateBox(0, s.scale[0], s.scale[1], s.scale[2]*12.0);
		dGeomSetBody(s.geomid, s.bodyid);
		const dReal* q = s.rotation.get_v();
		dMatrix3 r;
		dQtoR(q, r);
		Matrix33 mr(r[0], r[1], r[2], r[4], r[5], r[6], r[8], r[9], r[10]);
		Vector3 v1(0.0, 0.0, s.scale[2]*6.0);
		Vector3 v2(mr * v1);
		dGeomSetPosition(s.geomid, s.translation[0]+v2[0], s.translation[1]+v2[1], s.translation[2]+v2[2]);
		dGeomSetQuaternion(s.geomid, s.rotation.get_v());
		dSpaceAdd(m_space, s.geomid);
		*/
	}
	else if (g_params.model == MODEL_SPIDER)
	{
		m_model.load("spider.x3d", m_world, m_space);
		//dWorldSetERP(m_world, 0.2);
		//dWorldSetCFM(m_world, 0.00001);
		m_rot_x = 15.0;
		m_rot_z = -95.0;
		m_cam_dist = 3.0;
	}
	else if (g_params.model == MODEL_BOARD)
	{
		m_model.load("snowboard.x3d", m_world, m_space);
		m_rot_x = 30.0;
		m_rot_z = 0.0;
		m_cam_dist = 9.0;
	}
	else
		assert(0);
	m_model.reset();
	m_pickspring = false;
}

void Simulator::generate_ground(const Vector3& translation, const Vector4& rotation, const Vector3& scale, real rand_up, real pos_z)
{
	srand(123001);
	Solid main_ground;
	main_ground.name = "main_ground";
	main_ground.type = SOLID_BOX;
	main_ground.scale = scale;
	main_ground.rotation = rotation;
	main_ground.rotation.normalize();
	main_ground.translation = translation;
	m_model.add_geom(main_ground, m_space);
	m_ground.push_back(main_ground.geomid);
	for (int i = 0; i < 30; i++)
	{
		Solid sphere;
		sphere.name = "sphere";
		sphere.type = SOLID_SPHERE;
		sphere.scale = Vector3(20.0, 20.0, 20.0);
		sphere.rotation = Vector4(0.0, 1.0, 0.0, 0.0);
		real randx = ((real)rand() / RAND_MAX - 0.5) * (main_ground.scale.x - 5.0); 
		real randy = ((real)rand() / RAND_MAX - 0.5) * (main_ground.scale.y - 5.0);
		real randz = ((real)rand() / RAND_MAX - 0.5) * rand_up;
		Vector3 pos(randx, randy, randz);
		dMatrix3 r;
		dQtoR(main_ground.rotation.get_v(), r);
		Matrix33 m(r[0], r[1], r[2], r[4], r[5], r[6], r[8], r[9], r[10]);
		pos = m * pos;
		pos += main_ground.translation;
		pos.z += pos_z;
		sphere.translation = pos;
		m_model.add_geom(sphere, m_space);
		m_ground.push_back(sphere.geomid);
	}
}

void Simulator::init_opengl()
{	
#if RENDER
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);
	glShadeModel(GL_FLAT);
	glColorMaterial (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);	
#endif
}

void Simulator::reshape(int width, int height)
{
#if RENDER
	m_width = width;
	m_height = height;
	m_aspect = (GLfloat)width / (GLfloat)height;
	
	glViewport(0, 0, (GLint)width, (GLint)height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(40.0, m_aspect, 1.0, 100.0);	
#endif
}

bool Simulator::is_ground(dGeomID g)
{
	for (int i = 0; i < m_ground.size(); i++)
		if (g == m_ground[i])
			return true;
	return false;
}

void Simulator::near_callback(void *data, dGeomID o1, dGeomID o2)
{
	Simulator* sim = (Simulator*)data;

	bool gr1 = sim->is_ground(o1);
	bool gr2 = sim->is_ground(o2);
#if !COLLIDE_WITH_SELF
	if (!(gr1 ^ gr2))
		return;
#endif
	
	const int N = 40;
	dContact contact[N];
	int n = dCollide(o1, o2, N, &contact[0].geom, sizeof(dContact));
	assert(n < N);
	if (n > 0)
	{
		if (g_params.model == MODEL_BOARD || g_params.model == MODEL_SNOWBOARDER)
		{
			dBodyID boardid = sim->m_model.get_board_bodyid();
			const dReal* rot = dBodyGetRotation(boardid);
			Vector3 board_dir(rot[0], rot[4], rot[8]);
			assert(board_dir.length2() > 0.9 && board_dir.length2() < 1.1);
			for (int i = 0; i < n; i++) 
			{
				Vector3 normal(contact[i].geom.normal);
				Vector3 dirtemp = cross(board_dir, normal);
				Vector3 dir1 = cross(normal, dirtemp);
				dir1.normalize();
				contact[i].fdir1[0] = dir1.x;
				contact[i].fdir1[1] = dir1.y;
				contact[i].fdir1[2] = dir1.z;
				contact[i].surface.mode = dContactMu2 | dContactFDir1 | dContactApprox1 | dContactSoftERP | dContactSoftCFM;
				contact[i].surface.mu = 0.1;
				contact[i].surface.mu2 = 0.5;
				contact[i].surface.soft_erp = 0.3;
				contact[i].surface.soft_cfm = 0.0005;
				dJointID c = dJointCreateContact(sim->m_world, sim->m_contactgroup, &contact[i]);
				dBodyID b1 = dGeomGetBody(contact[i].geom.g1);
				dBodyID b2 = dGeomGetBody(contact[i].geom.g2);
				dJointAttach(c, b1, b2);
				if (g_params.is_annealing)
				{
					if (gr1 && (b2 != sim->m_model.get_board_bodyid()) || gr2 && (b1 != sim->m_model.get_board_bodyid()))
						the_annealing->set_abort();
				}
			}
		}
		else if (g_params.model == MODEL_SPIDER)
		{
			for (int i = 0; i < n; i++) 
			{
				//contact[i].surface.mode = dContactSlip1 | dContactSlip2 | dContactSoftERP | dContactSoftCFM | dContactApprox1;
				contact[i].surface.mode = dContactBounce | dContactApprox1;
				contact[i].surface.mu = 0.5;
				contact[i].surface.bounce = 0.2;
				/*
				contact[i].surface.slip1 = 0.1;
				contact[i].surface.slip2 = 0.1;
				contact[i].surface.soft_erp = 0.5;
				contact[i].surface.soft_cfm = 0.3;
				*/
				dJointID c = dJointCreateContact(sim->m_world, sim->m_contactgroup, &contact[i]);
				dBodyID b1 = dGeomGetBody(contact[i].geom.g1);
				dBodyID b2 = dGeomGetBody(contact[i].geom.g2);
				dJointAttach(c, b1, b2);
				if (g_params.is_annealing)
				{
					if (b1 == sim->m_model.get_head_bodyid() || b2 == sim->m_model.get_head_bodyid())
						the_annealing->set_abort();
				}
			}
		}
		else if (g_params.model == MODEL_CYCLIST)
		{
			for (int i = 0; i < n; i++) 
			{
				//contact[i].surface.mode = dContactSlip1 | dContactSlip2 | dContactSoftERP | dContactSoftCFM | dContactApprox1;
				contact[i].surface.mode = dContactApprox1;
				contact[i].surface.mu = 2.0;
				dJointID c = dJointCreateContact(sim->m_world, sim->m_contactgroup, &contact[i]);
				dBodyID b1 = dGeomGetBody(contact[i].geom.g1);
				dBodyID b2 = dGeomGetBody(contact[i].geom.g2);
				dJointAttach(c, b1, b2);
				if (g_params.is_annealing)
				{
					bool wheel1touching = (b1 == sim->m_model.get_bicycle_wheel1id() || b2 == sim->m_model.get_bicycle_wheel1id());
					bool wheel2touching = (b1 == sim->m_model.get_bicycle_wheel2id() || b2 == sim->m_model.get_bicycle_wheel2id());
					if (!wheel1touching && !wheel2touching)
						the_annealing->set_abort();
				}
			}
		}
		else
			assert(0);
	}
}
void Simulator::step(real t)
{
	real ks = 1000.0;
	real kd = 4000.0;
	real pos[3];
	m_model.get_head_pos(pos);
	Vector3 model_pos(pos);
	Vector3 force_pos(0.0, 0.0, 0.0);
	if (g_params.model == MODEL_CYCLIST)
		force_pos = Vector3(0.1, 0.0, 0.5);
	real fpw[3];
	dBodyID head = m_model.get_head_bodyid();
	dBodyGetRelPointPos(head, force_pos.x, force_pos.y, force_pos.z, fpw);
	Vector3 force_pos_world(fpw[0], fpw[1], fpw[2]);
	if (m_pickspring)
	{
		Vector3 spring_dir(m_pick - force_pos_world);
		spring_dir *= ks;
		Vector3 model_vel = force_pos_world - m_prev_model_pos;
		model_vel *= kd;
		spring_dir -= model_vel;
		//cout << "force " << spring_dir.x << " " << spring_dir.y << " " << spring_dir.z << endl;
		dBodyAddForceAtRelPos(head, spring_dir.x, spring_dir.y, spring_dir.z, force_pos.x, force_pos.y, force_pos.z);
	}
	
	m_prev_model_pos = force_pos_world;
	
	m_model.step(t);
	m_timer.elapsed();
	dSpaceCollide(m_space, this, &(Simulator::near_callback));
	////real time_coll = m_timer.elapsed();
	dWorldStep(m_world, t);
	////real time_step = m_timer.elapsed();
    dJointGroupEmpty(m_contactgroup);
	////cout << "dSpaceCollide: " << time_coll << " dWorldStep: " << time_step << endl;
}

void Simulator::draw()
{
#if RENDER
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	/*
	float m[16] = {
		1.0, 0.0, 0.0, 0.0,
		0.0, 0.0, -1.0, 0.0,
		0.0, 1.0, 0.0, 0.0,
		0.0, 0.0, 0.0, 1.0 };
	glMultMatrixf(m);
	glTranslatef(0.0, m_cam_dist, 0.0);
	glRotatef(m_rot_x, 1.0, 0.0, 0.0);
	glRotatef(m_rot_z, 0.0, 0.0, 1.0);
	 */
	
	static GLfloat light_snowboarder[4] = {50.0f, -50.0f, 10.0f, 0.0f};
	static GLfloat light_spider[4] = {50.0f, 50.0f, 10.0f, 0.0f};

	if (g_params.model == MODEL_SNOWBOARDER || g_params.model == MODEL_BOARD || g_params.model == MODEL_CYCLIST)
	{
		gluLookAt(4.0, -m_cam_dist, 10.0, 7.0, 0.0, 0.0, 0.0, 0.0, 1.0);
		glLightfv(GL_LIGHT0, GL_POSITION, light_snowboarder);		
	}
	else if (g_params.model == MODEL_SPIDER)
	{
		gluLookAt(m_cam_dist, 1.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
		glLightfv(GL_LIGHT0, GL_POSITION, light_spider);
		glColor3f(0.3f, 0.5f, 0.5f);
		glBegin(GL_QUADS);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-20.0f, -10.0f, 0.0f);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(20.0f, -10.0f, 0.0f);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(20.0f, 10.0f, 0.0f);
		glNormal3f(0.0f, 0.0f, 1.0f);
		glVertex3f(-20.0f, 10.0f, 0.0f);
		glEnd();
	}
	else
		assert(0);

	glColor3f(1.0f, 1.0f, 1.0f);
	m_model.draw();
	//glColor3f(0.4f, 0.6f, 0.6f);
	//m_mesh.draw();
	
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_LINES);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(1.0f, 0.0f, 0.0f);
	glEnd();
	glColor3f(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(0.0f, 1.0f, 0.0f);
	glEnd();
	glColor3f(0.0, 0.0, 1.0);
	glBegin(GL_LINES);
	glVertex3f(0.0f, 0.0f, 0.0f);
	glVertex3f(0.0f, 0.0f, 1.0f);
	glEnd();
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
#endif
}

void Simulator::keypress(unsigned char key, int x, int y)
{
	if (key == 's')
		the_annealing->set_stop(true);
}

Vector3 Simulator::get_projection(int x, int y)
{
#if RENDER
	GLdouble fmv[16];
	GLdouble fpr[16];
	GLint ivp[4];
		
	glGetDoublev(GL_MODELVIEW_MATRIX, fmv);
	glGetDoublev(GL_PROJECTION_MATRIX, fpr);
	glGetIntegerv(GL_VIEWPORT, ivp);
	
	int winy = ivp[3] - (GLint)y - 1;
	GLdouble objx, objy, objz;
	
	GLdouble winz;
	if (g_params.model == MODEL_SPIDER)
		winz = 0.65;
	else
		winz = 0.96;

	gluUnProject((GLdouble)x, (GLdouble)winy, winz, fmv, fpr, ivp, &objx, &objy, &objz);
	
	return Vector3(objx, objy, objz);
#else
	return Vector3(0.0, 0.0, 0.0);
#endif
}

void Simulator::mouse_button(int button, int state, int x, int y)
{
#if RENDER
	if (state == GLUT_DOWN)
	{
		m_pick = get_projection(x, y);
		m_pickspring = true;
	}
	else if (state == GLUT_UP)
	{
		m_pickspring = false;
	}
	/*
	if (state == GLUT_DOWN)
	{
		m_prev_mouse_x = x;
		m_prev_mouse_y = y;
	}
	 */
#endif
}

#define MOUSE_ROT_FACTOR 1.0;

void Simulator::mouse_drag(int x, int y)
{
	/*
	int x_delta = x - m_prev_mouse_x;
	int y_delta = y - m_prev_mouse_y;
	m_rot_z += (real)x_delta * MOUSE_ROT_FACTOR;
	m_rot_x += (real)y_delta * MOUSE_ROT_FACTOR;
	m_prev_mouse_x = x;
	m_prev_mouse_y = y;
	 */
	m_pick = get_projection(x, y);
}

