//
//  main.m
//  alife
//
//  Created by Eduardo Poyart on 2/17/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#include "alife_Prefix.pch"
#include "common.h"

Simulator* the_sim;
Annealing* the_annealing;

Timer the_timer;
GlobalParams g_params;

void init()
{
	the_sim = new Simulator();
	the_sim->init();
	the_sim->init_opengl();
	the_timer.elapsed();
	if (g_params.is_annealing)
	{
		the_annealing = new Annealing(the_sim);
		the_annealing->init();
	}
}

void cb_draw()
{
	the_sim->draw();
#if RENDER
	glutSwapBuffers();
#endif
}

void cb_reshape(int width, int height)
{
	the_sim->reshape(width, height);
}

#define TIMER_MSECS (1000.0 / 60.0)

void cb_timer(int value)
{
	//real tdelta = the_timer.elapsed();	
	//the_sim->step(tdelta);
	the_sim->step(1.0 / 60.0);
#if RENDER
	glutPostRedisplay();
	glutTimerFunc(TIMER_MSECS, cb_timer, value);
#endif
}

#if RENDER
static int render_interval = 0;
static int render_display = 1;
#endif

void cb_annealing_idle()
{
	if (g_params.is_annealing)
		the_annealing->step();
#if RENDER
	render_interval++;
	if (render_interval == render_display)
	{
		render_interval = 0;
		glutPostRedisplay();
	}
#endif
}

void cb_key(unsigned char key, int x, int y)
{
	the_sim->keypress(key, x, y);
}

void cb_mouse(int button, int state, int x, int y)
{
	the_sim->mouse_button(button, state, x, y);
}

void cb_motion(int x, int y)
{
	the_sim->mouse_drag(x, y);
}

/*
    0  0
 0  0  0  0
 0  0  0  0
    0  0

 w: 3 x 4 x 4 = 48
 o: 3 x 4 = 12
 
 */
void unit_tests()
{
	NeuralNet nn(2, 2, 4, 2);
	int ps = nn.parameter_space();
	assert(ps == 60);
	for (int i = 0; i < ps; i++)
		cout << *nn.parameter(i) << " ";
	cout << endl;
	*nn.parameter(0) = 10.0;
	*nn.parameter(8) = 1.0;
	*nn.parameter(9) = 1.0;
	*nn.parameter(16) = 10.0;
	*nn.parameter(33) = 10.0;
	*nn.parameter(48) = -12.0;
	vector<real> input;
	vector<real> output;
	
	input.resize(2);
	output.resize(2);
	input[0] = 1.0;
	input[1] = 0.0;
	nn.compute(input, output);
	cout << "Res: ";
	for (int i = 0; i < 2; i++)
		cout << output[i] << " ";
	cout << endl;
}

int main(int argc, char *argv[])
{
	g_params.model = MODEL_CYCLIST;
	g_params.is_annealing = false;
#if PROOF_OF_CONCEPT
	assert(g_params.is_annealing == false);
#endif
#if UNIT_TESTS
	unit_tests();
	exit(0);
#endif
#if RENDER
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize(740, 500);
	glutCreateWindow("Artificial Life");
#endif
	
	init();

#if RENDER
	glutDisplayFunc(cb_draw);
	glutReshapeFunc(cb_reshape);
	if (g_params.is_annealing)
	{
		glutIdleFunc(cb_annealing_idle);
		glutKeyboardFunc(cb_key);
	}
	else
	{
		if (g_params.model == MODEL_SNOWBOARDER || g_params.model == MODEL_CYCLIST)
			the_sim->get_model().get_nn()->load("nn.txt");
		else if (g_params.model == MODEL_SPIDER)
			the_sim->get_model().get_fj()->load("fj.txt");
		else if (g_params.model == MODEL_BOARD) {}
		else
			assert(0);
		glutKeyboardFunc(cb_key);
		//glutSpecialFunc(cb_special);
		glutMouseFunc(cb_mouse);
		glutMotionFunc(cb_motion);
		glutTimerFunc(TIMER_MSECS, cb_timer, 0);
	}
	glutMainLoop();
#else // RENDER
	if (g_params.is_annealing)
	{
		while(1)
			cb_annealing_idle();
	}
	else
	{
		cout << "Nothing to do - no render and no annealing." << endl;
	}
#endif // RENDER
	return 0;
}
